import itertools
from typing import Any, Dict, List

import numpy as np
import pandas as pd

import gobbli.io
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import PredictMixin, TrainMixin
from gobbli.util import multilabel_to_indicator_df, pred_prob_to_pred_label


class MajorityClassifier(BaseModel, TrainMixin, PredictMixin):
    """
    Simple classifier that returns the majority class from the training set.

    Useful for ensuring user code works with the gobbli input/output format
    without having to build a time-consuming model.
    """

    def init(self, params: Dict[str, Any]):
        self.majority_class: Any = None

    def _build(self):
        """
        No build step required for this model.
        """

    def _make_pred_df(self, labels: List[str], size: int) -> pd.DataFrame:
        return pd.DataFrame(
            {label: 1 if label == self.majority_class else 0 for label in labels},
            index=range(size),
        )

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        """
        Determine the majority class.
        """
        if train_input.multilabel:
            train_labels: List[str] = list(
                itertools.chain.from_iterable(train_input.y_train_multilabel)
            )
        else:
            train_labels = train_input.y_train_multiclass

        unique_values, value_counts = np.unique(train_labels, return_counts=True)
        self.majority_class = unique_values[value_counts.argmax(axis=0)]

        labels = train_input.labels()
        y_train_pred_proba = self._make_pred_df(labels, len(train_input.y_train))
        y_valid_pred_proba = self._make_pred_df(labels, len(train_input.y_valid))

        if train_input.multilabel:
            y_train_indicator = multilabel_to_indicator_df(
                train_input.y_train_multilabel, labels
            )
            train_loss = (
                (y_train_pred_proba.subtract(y_train_indicator)).abs().to_numpy().sum()
            )

            y_valid_indicator = multilabel_to_indicator_df(
                train_input.y_valid_multilabel, labels
            )
            valid_loss = (
                (y_valid_pred_proba.subtract(y_valid_indicator)).abs().to_numpy().sum()
            )
            valid_accuracy = valid_loss / (
                y_valid_pred_proba.shape[0] * y_valid_pred_proba.shape[1]
            )
        else:
            y_train_pred = pred_prob_to_pred_label(y_train_pred_proba)
            train_loss = np.sum(y_train_pred != train_input.y_train_multiclass)

            y_valid_pred = pred_prob_to_pred_label(y_valid_pred_proba)
            valid_loss = np.sum(y_valid_pred != train_input.y_valid_multiclass)
            valid_accuracy = valid_loss / len(y_valid_pred)

        return gobbli.io.TrainOutput(
            valid_loss=valid_loss,
            valid_accuracy=valid_accuracy,
            train_loss=train_loss,
            labels=train_input.labels(),
            multilabel=train_input.multilabel,
        )

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:
        """
        Predict based on our learned majority class.
        """
        pred_proba_df = self._make_pred_df(predict_input.labels, len(predict_input.X))

        return gobbli.io.PredictOutput(y_pred_proba=pred_proba_df)



from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import altair as alt
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    f1_score,
    precision_score,
    recall_score,
)

from gobbli.util import (
    as_multiclass,
    as_multilabel,
    escape_line_delimited_text,
    is_multilabel,
    multilabel_to_indicator_df,
    pred_prob_to_pred_label,
    pred_prob_to_pred_multilabel,
    truncate_text,
)


@dataclass
class ClassificationError:
    """
    Describes an error in classification.  Reports the original text,
    the true label, and the predicted probability.

    Args:
      X: The original text.
      y_true: The true label(s).
      y_pred_proba: The model predicted probability for each class.
    """

    X: str
    y_true: Union[str, List[str]]
    y_pred_proba: Dict[str, float]

    @property
    def y_pred(self) -> str:
        """
        Returns:
          The class with the highest predicted probability for this observation.
        """
        return max(self.y_pred_proba, key=lambda k: self.y_pred_proba[k])

    def y_pred_multilabel(self, threshold: float = 0.5) -> List[str]:
        """
        Args:
          threshold: The predicted probability threshold for predictions

        Returns:
          The predicted labels for this observation (predicted probability greater than
          the given threshold)
        """
        return pred_prob_to_pred_multilabel(self.y_pred_proba, threshold)


MetricFunc = Callable[[Sequence[str], pd.DataFrame], float]
"""
A function used to calculate some metric.  It should accept a sequence of true labels (y_true)
and a dataframe of shape (n_samples, n_classes) containing predicted probabilities; it should
output a real number.
"""


DEFAULT_METRICS: Dict[str, MetricFunc] = {
    "Weighted F1 Score": lambda y_true, y_pred: f1_score(
        y_true, y_pred, average="weighted"
    ),
    "Weighted Precision Score": lambda y_true, y_pred: precision_score(
        y_true, y_pred, average="weighted"
    ),
    "Weighted Recall Score": lambda y_true, y_pred: recall_score(
        y_true, y_pred, average="weighted"
    ),
    "Accuracy": lambda y_true, y_pred: accuracy_score(y_true, y_pred),
}
"""
The default set of metrics to evaluate classification models with.  Users may want to extend
this.
"""


@dataclass
class ClassificationEvaluation:
    """
    Provides several methods for evaluating the results from a classification problem.

    Args:
      labels: The set of unique labels in the dataset.
      X: The list of texts that were classified.
      y_true: The true labels for the dataset.
      y_pred_proba: A dataframe containing a row for each observation in X and a
        column for each label in the training data.  Cells are predicted probabilities.
    """

    labels: List[str]
    X: List[str]
    y_true: Union[List[str], List[List[str]]]
    y_pred_proba: pd.DataFrame
    metric_funcs: Optional[Dict[str, Callable[[Sequence, Sequence], float]]] = None

    def __post_init__(self):
        if not len(self.y_true) == self.y_pred_proba.shape[0]:
            raise ValueError(
                "y_true and y_pred_proba must have the same number of observations"
            )

        self.multilabel = is_multilabel(self.y_true)

    @property
    def y_true_multiclass(self) -> List[str]:
        return as_multiclass(self.y_true, self.multilabel)

    @property
    def y_true_multilabel(self) -> pd.DataFrame:
        return multilabel_to_indicator_df(
            as_multilabel(self.y_true, self.multilabel), self.labels
        )

    @property
    def y_pred_multiclass(self) -> List[str]:
        """
        Returns:
          Predicted class for each observation (assuming multiclass context).
        """
        return pred_prob_to_pred_label(self.y_pred_proba)

    @property
    def y_pred_multilabel(self) -> pd.DataFrame:
        """
        Returns:
          Indicator dataframe containing a 0 if each label wasn't predicted and 1 if
          it was for each observation.
        """
        return pred_prob_to_pred_multilabel(self.y_pred_proba).astype("int")

    def metrics(self) -> Dict[str, float]:
        """
        Returns:
          A dictionary containing various metrics of model performance on the test dataset.
        """
        metric_funcs = self.metric_funcs
        if metric_funcs is None:
            metric_funcs = DEFAULT_METRICS

        if self.multilabel:
            y_true: Union[List[str], pd.DataFrame] = self.y_true_multilabel
            y_pred: Union[List[str], pd.DataFrame] = self.y_pred_multilabel
        else:
            y_true = self.y_true_multiclass
            y_pred = self.y_pred_multiclass

        return {
            name: metric_func(y_true, y_pred)
            for name, metric_func in metric_funcs.items()
        }

    def metrics_report(self) -> str:
        """
        Returns:
          A nicely formatted human-readable report describing metrics of model performance
          on the test dataset.
        """
        metric_string = "\n".join(
            f"{name}: {metric}" for name, metric in self.metrics().items()
        )

        if self.multilabel:
            y_true: Union[pd.DataFrame, List[str]] = self.y_true_multilabel
            y_pred: Union[pd.DataFrame, List[str]] = self.y_pred_multilabel
            # Since these are indicator dataframes, the "labels" are indices
            labels: Union[List[str], List[int]] = list(range(len(self.labels)))
        else:
            y_true = self.y_true_multiclass
            y_pred = self.y_pred_multiclass
            # Since these are lists of labels, the "labels" are the strings themselves
            labels = self.labels

        return (
            "Metrics:\n"
            "--------\n"
            f"{metric_string}\n\n"
            "Classification Report:\n"
            "----------------------\n"
            f"{classification_report(y_true, y_pred, labels=labels, target_names=self.labels)}\n"
        )

    def plot(self, sample_size: Optional[int] = None) -> alt.Chart:
        """
        Args:
          sample_size: Optional number of points to sample for the plot.  Unsampled
            plots may be difficult to save due to their size.

        Returns:
          An Altair chart visualizing predicted probabilities and true classes to visually identify
          where errors are being made.
        """
        # Since multilabel is a generalization of the multiclass paradigm, implement
        # this visualization the same for multiclass and multilabel using the multilabel
        # format
        pred_prob_df = self.y_pred_proba
        true_df = self.y_true_multilabel

        if sample_size is not None:
            # Avoid errors due to sample being larger than the population if the number
            # of observations is smaller than the sample size
            pred_prob_df = pred_prob_df.sample(
                n=min(sample_size, pred_prob_df.shape[0])
            )
            true_df = true_df.iloc[pred_prob_df.index]

        charts = []

        if self.multilabel:
            legend_label = "Has Label"
        else:
            legend_label = "Belongs to Class"

        for label in self.labels:
            # Plot the predicted probabilities for given label for all observations
            plot_df = (
                pred_prob_df[[label]]
                .rename({label: "Predicted Probability"}, axis="columns")
                .join(
                    true_df[[label]]
                    .astype("bool")
                    .rename({label: legend_label}, axis="columns")
                )
            )

            charts.append(
                alt.layer(
                    alt.Chart(plot_df, title=label, height=40)
                    .mark_circle(size=8)
                    .encode(
                        x=alt.X(
                            "Predicted Probability",
                            type="quantitative",
                            title=None,
                            scale=alt.Scale(domain=(0.0, 1.0)),
                        ),
                        y=alt.Y(
                            "jitter",
                            type="quantitative",
                            title=None,
                            axis=alt.Axis(
                                values=[0], ticks=True, grid=False, labels=False
                            ),
                            scale=alt.Scale(),
                        ),
                        color=alt.Color(legend_label, type="nominal"),
                    )
                    .transform_calculate(
                        # Generate Gaussian jitter with a Box-Muller transform
                        jitter="sqrt(-2*log(random()))*cos(2*PI*random())/32"
                    )
                    .properties(height=40)
                )
            )
        return alt.vconcat(*charts)

    def errors_for_label(self, label: str, k: int = 10):
        """
        Output the biggest mistakes for the given class by the classifier

        Args:
          label: The label to return errors for.
          k: The number of results to return for each of false positives and false negatives.

        Returns:
          A 2-tuple.  The first element is a list of the top ``k`` false positives, and the
          second element is a list of the top ``k`` false negatives.
        """
        pred_label = self.y_pred_multilabel[label].astype("bool")
        true_label = self.y_true_multilabel[label].astype("bool")

        # Order false positives/false negatives by the degree of the error;
        # i.e. we want the false positives with highest predicted probability first
        # and false negatives with lowest predicted probability first
        # Take the top `k` of each
        false_positives = (
            self.y_pred_proba.loc[pred_label & ~true_label]
            .sort_values(by=label, ascending=False)
            .iloc[:k]
        )
        false_negatives = (
            self.y_pred_proba.loc[~pred_label & true_label]
            .sort_values(by=label, ascending=True)
            .iloc[:k]
        )

        def create_classification_errors(
            y_pred_proba: pd.DataFrame,
        ) -> List[ClassificationError]:
            classification_errors = []
            for ndx, row in y_pred_proba.iterrows():
                classification_errors.append(
                    ClassificationError(
                        X=self.X[ndx],
                        y_true=self.y_true[ndx],
                        y_pred_proba=row.to_dict(),
                    )
                )
            return classification_errors

        return (
            create_classification_errors(false_positives),
            create_classification_errors(false_negatives),
        )

    def errors(
        self, k: int = 10
    ) -> Dict[str, Tuple[List[ClassificationError], List[ClassificationError]]]:
        """
        Output the biggest mistakes for each class by the classifier.

        Args:
          k: The number of results to return for each of false positives and false negatives.

        Returns:
          A dictionary whose keys are label names and values are 2-tuples.  The first
          element is a list of the top ``k`` false positives, and the second element is a list
          of the top ``k`` false negatives.
        """
        errors = {}
        for label in self.labels:
            errors[label] = self.errors_for_label(label, k=k)

        return errors

    def errors_report(self, k: int = 10) -> str:
        """
        Args:
          k: The number of results to return for each of false positives and false negatives.

        Returns:
          A nicely-formatted human-readable report describing the biggest mistakes made by
          the classifier for each class.
        """
        errors = self.errors(k=k)
        output = "Errors Report\n" "------------\n\n"

        for label, (false_positives, false_negatives) in errors.items():

            def make_errors_str(errors: List[ClassificationError]) -> str:
                if self.multilabel:
                    return "\n".join(
                        (
                            f"Correct Value: {label in e.y_true}\n"
                            f"Predicted Probability: {e.y_pred_proba[label]}"
                            f"Text: {truncate_text(escape_line_delimited_text(e.X), 500)}\n"
                        )
                        for e in errors
                    )
                else:
                    return "\n".join(
                        (
                            f"True Class: {e.y_true}\n"
                            f"Predicted Class: {e.y_pred} (Probability: {e.y_pred_proba[e.y_pred]})\n"
                            f"Text: {truncate_text(escape_line_delimited_text(e.X), 500)}\n"
                        )
                        for e in errors
                    )

            false_positives_str = make_errors_str(false_positives)
            if len(false_positives_str) == 0:
                false_positives_str = "None"
            false_negatives_str = make_errors_str(false_negatives)
            if len(false_negatives_str) == 0:
                false_negatives_str = "None"

            header_name = "CLASS" if self.multilabel else "LABEL"

            output += (
                " -------\n"
                f"| {header_name}: {label}\n"
                " -------\n\n"
                "False Positives\n"
                "***************\n\n"
                f"{false_positives_str}\n\n"
                "False Negatives\n"
                "***************\n\n"
                f"{false_negatives_str}\n\n"
            )

        return output



# Copyright (c) Microsoft. All rights reserved.
import tqdm
import unicodedata

PAD = 'PADPAD'
UNK = 'UNKUNK'
STA= 'BOSBOS'
END = 'EOSEOS'

PAD_ID = 0
UNK_ID = 1
STA_ID = 2
END_ID = 3

class Vocabulary(object):
    INIT_LEN = 4
    def __init__(self, neat=False):
        self.neat = neat
        if not neat:
            self.tok2ind = {PAD: PAD_ID, UNK: UNK_ID, STA: STA_ID, END: END_ID}
            self.ind2tok = {PAD_ID: PAD, UNK_ID: UNK, STA_ID: STA, END_ID:END}
        else:
            self.tok2ind = {}
            self.ind2tok = {}

    def __len__(self):
        return len(self.tok2ind)

    def __iter__(self):
        return iter(self.tok2ind)

    def __contains__(self, key):
        if type(key) == int:
            return key in self.ind2tok
        elif type(key) == str:
            return key in self.tok2ind

    def __getitem__(self, key):
        if type(key) == int:
            return self.ind2tok.get(key, -1) if self.neat else self.ind2tok.get(key, UNK)
        if type(key) == str:
            return self.tok2ind.get(key, None) if self.neat else self.tok2ind.get(key,self.tok2ind.get(UNK))

    def __setitem__(self, key, item):
        if type(key) == int and type(item) == str:
            self.ind2tok[key] = item
        elif type(key) == str and type(item) == int:
            self.tok2ind[key] = item
        else:
            raise RuntimeError('Invalid (key, item) types.')

    def add(self, token):
        if token not in self.tok2ind:
            index = len(self.tok2ind)
            self.tok2ind[token] = index
            self.ind2tok[index] = token

    def get_vocab_list(self, with_order=True):
        if with_order:
            words = [self[k] for k in range(0, len(self))]
        else:
            words = [k for k in self.tok2ind.keys()
                      if k not in {PAD, UNK, STA, END}]
        return words

    def toidx(self, tokens):
        return [self[tok] for tok in tokens]

    def copy(self):
        """Deep copy
        """
        new_vocab = Vocabulary(self.neat)
        for w in self:
            new_vocab.add(w)
        return new_vocab

    def build(words, neat=False):
        vocab = Vocabulary(neat)
        for w in words: vocab.add(w)
        return vocab



import pandas as pd

from gobbli.dataset.newsgroups import NewsgroupsDataset


def test_load_newsgroups(tmp_gobbli_dir):
    ds = NewsgroupsDataset.load()

    X_train = ds.X_train()
    X_test = ds.X_test()

    y_train = ds.y_train()
    y_test = ds.y_test()

    assert len(X_train) == 11314
    assert len(y_train) == 11314
    assert len(X_test) == 7532
    assert len(y_test) == 7532

    assert len(pd.unique(y_train)) == 20
    assert len(pd.unique(y_test)) == 20

    # Ensure these objects pass validation
    ds.train_input()
    ds.predict_input()



"""
Mixins which can be applied to classes derived from BaseModel.
"""
from abc import ABCMeta, abstractmethod
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Callable, Optional, cast

import gobbli.io
from gobbli.model.context import ContainerTaskContext
from gobbli.util import format_duration, generate_uuid, write_metadata


def _run_task(
    task_func: Callable[[Any, ContainerTaskContext], Any],
    task_input: gobbli.io.TaskIO,
    root_dir: Path,
    dir_name: Optional[str] = None,
) -> gobbli.io.TaskIO:
    """
    Run a task function that generates some output.  Can create a unique id
    to name the directory storing the input/output or use a user-provided name.
    Generate a context object to pass to the task.
    """
    if dir_name is None:
        task_id = generate_uuid()
        task_root_dir = root_dir / task_id
    else:
        task_root_dir = root_dir / dir_name

    if task_root_dir.exists():
        raise ValueError(
            f"Directory '{task_root_dir}' already exists.  Supply a different `dir_name`."
        )
    context = ContainerTaskContext(task_root_dir=task_root_dir)

    write_metadata(
        task_input.metadata(),
        context.host_input_dir / gobbli.io.TaskIO._METADATA_FILENAME,
    )

    task_output = cast(gobbli.io.TaskIO, task_func(task_input, context))

    write_metadata(
        task_output.metadata(),
        context.host_output_dir / gobbli.io.TaskIO._METADATA_FILENAME,
    )

    return task_output


class TrainMixin(metaclass=ABCMeta):
    """
    Apply to a model which can be trained in some way.
    """

    @abstractmethod
    def data_dir(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def logger(self):
        raise NotImplementedError

    def train_dir(self) -> Path:
        """
        The directory to be used for data related to training (data files, etc).

        Returns:
          Path to the training data directory.
        """
        return self.data_dir() / "train"

    def train(
        self, train_input: gobbli.io.TrainInput, train_dir_name: Optional[str] = None
    ) -> gobbli.io.TrainOutput:
        """
        Trains a model using params in the given :obj:`gobbli.io.TrainInput`.
        The training process varies depending on the model, but in general, it includes
        the following steps:

        - Update weights using the training dataset
        - Evaluate performance on the validation dataset.
        - Repeat for a number of epochs.
        - When finished, report loss/accuracy and return the trained weights.

        Args:
          train_input: Contains various parameters needed to determine
            how to train and what data to train on.
          train_dir_name: Optional name to store training input and output
            under.  The directory is always created under the model's
            :meth:`data_dir<gobbli.model.base.BaseModel.data_dir>`.  If a name is not given,
            a unique name is generated via a UUID.  If a name is given, that
            directory must not already exist.

        Returns:
          Output of training.
        """
        self.logger.info("Starting training.")
        start = timer()
        train_output = cast(
            gobbli.io.TrainOutput,
            _run_task(self._train, train_input, self.train_dir(), train_dir_name),
        )
        end = timer()
        for log_line in (
            f"Training finished in {format_duration(end - start)}.",
            "RESULTS:",
            f"  Validation loss: {train_output.valid_loss}",
            f"  Validation accuracy: {train_output.valid_accuracy}",
            f"  Training loss: {train_output.train_loss}",
        ):
            self.logger.info(log_line)

        return train_output

    @abstractmethod
    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        raise NotImplementedError


class PredictMixin(metaclass=ABCMeta):
    """
    Apply to a model which can be used to predict on new data.
    """

    @abstractmethod
    def data_dir(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def logger(self):
        raise NotImplementedError

    def predict_dir(self) -> Path:
        """
        The directory to be used for data related to prediction (weights, predictions, etc)

        Returns:
          Path to the prediction data directory.
        """
        return self.data_dir() / "predict"

    def predict(
        self,
        predict_input: gobbli.io.PredictInput,
        predict_dir_name: Optional[str] = None,
    ) -> gobbli.io.PredictOutput:
        """
        Runs prediction on new data using params containing in the given :obj:`gobbli.io.PredictInput`.

        Args:
          predict_input:  Contains various parameters needed to determine how to run prediction
            and what data to predict for.
          predict_dir_name: Optional name to store prediction input and output
            under.  The directory is always created under the model's
            :meth:`data_dir<gobbli.model.base.BaseModel.data_dir>`.  If a name is not given,
            a unique name is generated via a UUID.  If a name is given, that
            directory must not already exist.
        """
        self.logger.info("Starting prediction.")
        start = timer()
        predict_output = cast(
            gobbli.io.PredictOutput,
            _run_task(
                self._predict, predict_input, self.predict_dir(), predict_dir_name
            ),
        )
        end = timer()

        self.logger.info(f"Prediction finished in {format_duration(end - start)}.")

        return predict_output

    @abstractmethod
    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:
        raise NotImplementedError


class EmbedMixin(metaclass=ABCMeta):
    """
    Apply to a model which can be used to generate embeddings from data.
    """

    @abstractmethod
    def data_dir(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def logger(self):
        raise NotImplementedError

    def embed_dir(self) -> Path:
        """
        The directory to be used for data related to embedding (weights, embeddings, etc)

        Returns:
          Path to the embedding data directory.
        """
        return self.data_dir() / "embed"

    def embed(
        self, embed_input: gobbli.io.EmbedInput, embed_dir_name: Optional[str] = None
    ) -> gobbli.io.EmbedOutput:
        """
        Generates embeddings using a model and the params in the given :obj:`gobbli.io.EmbedInput`.

        Args:
          embed_input: Contains various parameters needed to determine
            how to generate embeddings and what data to generate embeddings for.
          embed_dir_name: Optional name to store embedding input and output
            under.  The directory is always created under the model's
            :meth:`data_dir<gobbli.model.base.BaseModel.data_dir>`.  If a name is not given,
            a unique name is generated via a UUID.  If a name is given, that
            directory must not already exist.

        Returns:
          Output of training.
        """
        self.logger.info("Generating embeddings.")
        start = timer()
        embed_output = cast(
            gobbli.io.EmbedOutput,
            _run_task(self._embed, embed_input, self.embed_dir(), embed_dir_name),
        )
        end = timer()

        self.logger.info(
            f"Embedding generation finished in {format_duration(end - start)}."
        )

        return embed_output

    @abstractmethod
    def _embed(
        self, embed_input: gobbli.io.EmbedInput, context: ContainerTaskContext
    ) -> gobbli.io.EmbedOutput:
        raise NotImplementedError



import pytest

import gobbli.io
from gobbli.dataset.newsgroups import NewsgroupsDataset
from gobbli.dataset.trivial import TrivialDataset
from gobbli.model.bert import BERT
from gobbli.model.fasttext import FastText
from gobbli.model.random import RandomEmbedder
from gobbli.model.sklearn import TfidfEmbedder
from gobbli.model.spacy import SpaCyModel
from gobbli.model.transformer import Transformer
from gobbli.model.use import USE
from gobbli.test.util import model_test_dir, skip_if_low_resource, validate_checkpoint


def check_embed_output(
    embed_input, embed_output, expected_dimensionality=None, max_seq_length=None
):
    """
    Verify some information regarding the embedding output.
    User can optionally pass expected values of the max sequence
    length and dimensionality, if they're known.
    If not, we'll just verify the dimensionality is larger than 0.
    """
    embeddings = embed_output.X_embedded
    assert len(embeddings) == len(embed_input.X)

    if embed_input.pooling == gobbli.io.EmbedPooling.NONE:
        for embedding, tokens in zip(embeddings, embed_output.embed_tokens):
            assert embedding.ndim == 2

            expected_length = len(tokens)
            if max_seq_length is not None and len(tokens) > max_seq_length:
                expected_length = max_seq_length

            assert embedding.shape[0] == expected_length
            if expected_dimensionality is not None:
                assert embedding.shape[1] == expected_dimensionality
            else:
                assert embedding.shape[1] > 0
    else:
        for embedding in embeddings:
            assert embedding.ndim == 1
            if expected_dimensionality is not None:
                assert embedding.shape[0] == expected_dimensionality
            else:
                assert embedding.shape[0] > 0


@pytest.mark.parametrize("pooling", list(gobbli.io.EmbedPooling))
@pytest.mark.parametrize(
    "model_cls,dataset_cls,model_kwargs,train_kwargs,embed_kwargs",
    [
        (RandomEmbedder, TrivialDataset, {}, {}, {}),
        (RandomEmbedder, NewsgroupsDataset, {}, {}, {}),
        (
            BERT,
            TrivialDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"embed_batch_size": 1},
        ),
        (
            BERT,
            NewsgroupsDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 8},
            {"embed_batch_size": 32},
        ),
        (USE, TrivialDataset, {}, {}, {"embed_batch_size": 1}),
        (USE, NewsgroupsDataset, {}, {}, {"embed_batch_size": 32}),
        (
            FastText,
            TrivialDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"embed_batch_size": 1},
        ),
        (
            FastText,
            NewsgroupsDataset,
            {},
            {"num_train_epochs": 1, "train_batch_size": 32, "valid_batch_size": 8},
            {"embed_batch_size": 32},
        ),
        (
            Transformer,
            TrivialDataset,
            {"max_seq_length": 128},
            {"num_train_epochs": 1, "train_batch_size": 1, "valid_batch_size": 1},
            {"embed_batch_size": 1},
        ),
        (
            Transformer,
            NewsgroupsDataset,
            {"max_seq_length": 128},
            {"num_train_epochs": 1, "train_batch_size": 16, "valid_batch_size": 8},
            {"embed_batch_size": 32},
        ),
        (SpaCyModel, TrivialDataset, {"model": "en_core_web_sm"}, {}, {}),
        (SpaCyModel, NewsgroupsDataset, {"model": "en_core_web_sm"}, {}, {}),
        (TfidfEmbedder, TrivialDataset, {}, {}, {}),
        (TfidfEmbedder, NewsgroupsDataset, {}, {}, {}),
    ],
)
def test_embeddings(
    model_cls,
    dataset_cls,
    model_kwargs,
    train_kwargs,
    embed_kwargs,
    model_gpu_config,
    pooling,
    gobbli_dir,
    request,
):
    """
    Ensure embedding models train and generate embeddings appropriately
    across a few example datasets.
    """
    if (
        model_cls in (USE, FastText, TfidfEmbedder)
        and pooling == gobbli.io.EmbedPooling.NONE
    ):
        pytest.xfail(f"pooling is required for {model_cls.__name__}")

    # These combinations of model and dataset require a lot of memory
    if model_cls in (BERT, Transformer) and dataset_cls in (NewsgroupsDataset,):
        skip_if_low_resource(request.config)

    model = model_cls(
        data_dir=model_test_dir(model_cls),
        load_existing=True,
        **model_gpu_config,
        **model_kwargs,
    )

    model.build()
    ds = dataset_cls.load()

    embed_input = ds.embed_input(limit=50, pooling=pooling, **embed_kwargs)
    check_kwargs = {}
    if isinstance(model, RandomEmbedder):
        check_kwargs["expected_dimensionality"] = RandomEmbedder.DIMENSIONALITY
    if isinstance(model, Transformer):
        check_kwargs["max_seq_length"] = model_kwargs.get("max_seq_length")

    # For models which support generating embeddings without training
    if model_cls not in (FastText,):
        # Verify we can generate embeddings without a trained checkpoint
        embed_output = model.embed(embed_input)
        check_embed_output(embed_input, embed_output, **check_kwargs)

    # Only these models support training for embeddings
    if model_cls in (BERT, FastText, Transformer):
        # Verify embedding runs with a trained checkpoint
        train_output = model.train(ds.train_input(limit=50, **train_kwargs))
        assert train_output.valid_loss is not None
        assert train_output.train_loss is not None
        assert 0 <= train_output.valid_accuracy <= 1

        validate_checkpoint(model_cls, train_output.checkpoint)

        embed_input.checkpoint = train_output.checkpoint
        embed_output = model.embed(embed_input)
        check_embed_output(embed_input, embed_output, **check_kwargs)



# Copyright (c) Microsoft. All rights reserved.
import torch
import random
import torch.nn as nn
from torch.nn.utils import weight_norm
from torch.autograd import Variable
from torch.nn.parameter import Parameter
import torch.nn.functional as F
import numpy as np
from module.dropout_wrapper import DropoutWrapper
from module.similarity import FlatSimilarityWrapper, SelfAttnWrapper
from module.my_optim import weight_norm as WN

SMALL_POS_NUM=1.0e-30

def generate_mask(new_data, dropout_p=0.0, is_training=False):
    if not is_training: dropout_p = 0.0
    new_data = (1-dropout_p) * (new_data.zero_() + 1)
    for i in range(new_data.size(0)):
        one = random.randint(0, new_data.size(1)-1)
        new_data[i][one] = 1
    mask = Variable(1.0/(1 - dropout_p) * torch.bernoulli(new_data), requires_grad=False)
    return mask


class Classifier(nn.Module):
    def __init__(self, x_size, y_size, opt, prefix='decoder', dropout=None):
        super(Classifier, self).__init__()
        self.opt = opt
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(prefix), 0))
        else:
            self.dropout = dropout
        self.merge_opt = opt.get('{}_merge_opt'.format(prefix), 0)
        self.weight_norm_on = opt.get('{}_weight_norm_on'.format(prefix), False)

        if self.merge_opt == 1:
            self.proj = nn.Linear(x_size * 4, y_size)
        else:
            self.proj = nn.Linear(x_size * 2, y_size)

        if self.weight_norm_on:
            self.proj = weight_norm(self.proj)

    def forward(self, x1, x2, mask=None):
        if self.merge_opt == 1:
            x = torch.cat([x1, x2, (x1 - x2).abs(), x1 * x2], 1)
        else:
            x = torch.cat([x1, x2], 1)
        x = self.dropout(x)
        scores = self.proj(x)
        return scores

class SANClassifier(nn.Module):
    """Implementation of Stochastic Answer Networks for Natural Language Inference, Xiaodong Liu, Kevin Duh and Jianfeng Gao
    https://arxiv.org/abs/1804.07888
    """
    def __init__(self, x_size, h_size, label_size, opt={}, prefix='decoder', dropout=None):
        super(SANClassifier, self).__init__()
        if dropout is None:
            self.dropout = DropoutWrapper(opt.get('{}_dropout_p'.format(self.prefix), 0))
        else:
            self.dropout = dropout
        self.prefix = prefix
        self.query_wsum = SelfAttnWrapper(x_size, prefix='mem_cum', opt=opt, dropout=self.dropout)
        self.attn = FlatSimilarityWrapper(x_size, h_size, prefix, opt, self.dropout)
        self.rnn_type = '{}{}'.format(opt.get('{}_rnn_type'.format(prefix), 'gru').upper(), 'Cell')
        self.rnn =getattr(nn, self.rnn_type)(x_size, h_size)
        self.num_turn = opt.get('{}_num_turn'.format(prefix), 5)
        self.opt = opt
        self.mem_random_drop = opt.get('{}_mem_drop_p'.format(prefix), 0)
        self.mem_type = opt.get('{}_mem_type'.format(prefix), 0)
        self.weight_norm_on = opt.get('{}_weight_norm_on'.format(prefix), False)
        self.label_size = label_size
        self.dump_state = opt.get('dump_state_on', False)
        self.alpha = Parameter(torch.zeros(1, 1), requires_grad=False)
        if self.weight_norm_on:
            self.rnn = WN(self.rnn)

        self.classifier = Classifier(x_size, self.label_size, opt, prefix=prefix, dropout=self.dropout)

    def forward(self, x, h0, x_mask=None, h_mask=None):
        h0 = self.query_wsum(h0, h_mask)
        if type(self.rnn) is nn.LSTMCell:
            c0 = Variable(h0.new(h0.size()).zero_())
        scores_list = []
        for turn in range(self.num_turn):
            att_scores = self.attn(x, h0, x_mask)
            x_sum = torch.bmm(F.softmax(att_scores, 1).unsqueeze(1), x).squeeze(1)
            scores = self.classifier(x_sum, h0)
            scores_list.append(scores)
            # next turn
            if self.rnn is not None:
                h0 = self.dropout(h0)
                if type(self.rnn) is nn.LSTMCell:
                    h0, c0 = self.rnn(x_sum, (h0, c0))
                else:
                    h0 = self.rnn(x_sum, h0)
        if self.mem_type == 1:
            mask = generate_mask(self.alpha.data.new(x.size(0), self.num_turn), self.mem_random_drop, self.training)
            mask = [m.contiguous() for m in torch.unbind(mask, 1)]
            tmp_scores_list = [mask[idx].view(x.size(0), 1).expand_as(inp) * F.softmax(inp, 1) for idx, inp in enumerate(scores_list)]
            scores = torch.stack(tmp_scores_list, 2)
            scores = torch.mean(scores, 2)
            scores = torch.log(scores)
        else:
            scores = scores_list[-1]
        if self.dump_state:
            return scores, scores_list
        else:
            return scores



# Copyright (c) Microsoft. All rights reserved.
import logging
from time import gmtime, strftime
import sys

def create_logger(name, silent=False, to_disk=False, log_file=None):
    """Logger wrapper
    """
    # setup logger
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.propagate = False
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S')
    if not silent:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    if to_disk:
        log_file = log_file if log_file is not None else strftime("%Y-%m-%d-%H-%M-%S.log", gmtime())
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        log.addHandler(fh)
    return log



# Copyright (c) Microsoft. All rights reserved.
import argparse
import json
import os
import random
from datetime import datetime
from pprint import pprint

import numpy as np
import torch

from data_utils.glue_utils import submit, eval_model
from data_utils.label_map import DATA_META, GLOBAL_MAP, DATA_TYPE, DATA_SWAP, TASK_TYPE, generate_decoder_opt
from data_utils.log_wrapper import create_logger
from data_utils.utils import set_environment
from mt_dnn.batcher import BatchGen
from mt_dnn.model import MTDNNModel
from pytorch_pretrained_bert.modeling import BertModel
from pytorch_pretrained_bert.modeling import BertConfig

def model_config(parser):
    parser.add_argument('--update_bert_opt',  default=0, type=int)
    parser.add_argument('--multi_gpu_on', action='store_true')
    parser.add_argument('--mem_cum_type', type=str, default='simple',
                        help='bilinear/simple/defualt')
    parser.add_argument('--answer_num_turn', type=int, default=5)
    parser.add_argument('--answer_mem_drop_p', type=float, default=0.1)
    parser.add_argument('--answer_att_hidden_size', type=int, default=128)
    parser.add_argument('--answer_att_type', type=str, default='bilinear',
                        help='bilinear/simple/defualt')
    parser.add_argument('--answer_rnn_type', type=str, default='gru',
                        help='rnn/gru/lstm')
    parser.add_argument('--answer_sum_att_type', type=str, default='bilinear',
                        help='bilinear/simple/defualt')
    parser.add_argument('--answer_merge_opt', type=int, default=1)
    parser.add_argument('--answer_mem_type', type=int, default=1)
    parser.add_argument('--answer_dropout_p', type=float, default=0.1)
    parser.add_argument('--answer_weight_norm_on', action='store_true')
    parser.add_argument('--dump_state_on', action='store_true')
    parser.add_argument('--answer_opt', type=int, default=0, help='0,1')
    parser.add_argument('--label_size', type=str, default='3')
    parser.add_argument('--mtl_opt', type=int, default=0)
    parser.add_argument('--ratio', type=float, default=0)
    parser.add_argument('--mix_opt', type=int, default=0)
    parser.add_argument('--max_seq_len', type=int, default=512)
    parser.add_argument('--init_ratio', type=float, default=1)
    return parser

def data_config(parser):
    parser.add_argument('--log_file', default='mt-dnn-train.log', help='path for log file.')
    parser.add_argument("--init_checkpoint", default='mt_dnn/bert_model_base.pt', type=str)
    parser.add_argument('--data_dir', default='data/mt_dnn')
    parser.add_argument('--data_sort_on', action='store_true')
    parser.add_argument('--name', default='farmer')
    parser.add_argument('--train_datasets', default='mnli')
    parser.add_argument('--test_datasets', default='mnli_mismatched,mnli_matched')
    parser.add_argument('--pw_tasks', default='qnnli', type=str)
    return parser

def train_config(parser):
    parser.add_argument('--cuda', type=bool, default=torch.cuda.is_available(),
                        help='whether to use GPU acceleration.')
    parser.add_argument('--log_per_updates', type=int, default=500)
    parser.add_argument('--epochs', type=int, default=5)
    parser.add_argument('--batch_size', type=int, default=8)
    parser.add_argument('--batch_size_eval', type=int, default=8)
    parser.add_argument('--optimizer', default='adamax',
                        help='supported optimizer: adamax, sgd, adadelta, adam')
    parser.add_argument('--grad_clipping', type=float, default=0)
    parser.add_argument('--global_grad_clipping', type=float, default=1.0)
    parser.add_argument('--weight_decay', type=float, default=0)
    parser.add_argument('--learning_rate', type=float, default=5e-5)
    parser.add_argument('--momentum', type=float, default=0)
    parser.add_argument('--warmup', type=float, default=0.1)
    parser.add_argument('--warmup_schedule', type=str, default='warmup_linear')

    parser.add_argument('--vb_dropout', action='store_false')
    parser.add_argument('--dropout_p', type=float, default=0.1)
    parser.add_argument('--dropout_w', type=float, default=0.000)
    parser.add_argument('--bert_dropout_p', type=float, default=0.1)

    # EMA
    parser.add_argument('--ema_opt', type=int, default=0)
    parser.add_argument('--ema_gamma', type=float, default=0.995)

    # scheduler
    parser.add_argument('--have_lr_scheduler', dest='have_lr_scheduler', action='store_false')
    parser.add_argument('--multi_step_lr', type=str, default='10,20,30')
    parser.add_argument('--freeze_layers', type=int, default=-1)
    parser.add_argument('--embedding_opt', type=int, default=0)
    parser.add_argument('--lr_gamma', type=float, default=0.5)
    parser.add_argument('--bert_l2norm', type=float, default=0.0)
    parser.add_argument('--scheduler_type', type=str, default='ms', help='ms/rop/exp')
    parser.add_argument('--output_dir', default='checkpoint')
    parser.add_argument('--seed', type=int, default=2018,
                        help='random seed for data shuffling, embedding init, etc.')
    parser.add_argument('--task_config_path', type=str, default='configs/tasks_config.json')

    return parser

parser = argparse.ArgumentParser()
parser = data_config(parser)
parser = model_config(parser)
parser = train_config(parser)
args = parser.parse_args()

output_dir = args.output_dir
data_dir = args.data_dir
args.train_datasets = args.train_datasets.split(',')
args.test_datasets = args.test_datasets.split(',')
args.pw_tasks = list(set([pw for pw in args.pw_tasks.split(',') if len(pw.strip()) > 0]))
pprint(args)

os.makedirs(output_dir, exist_ok=True)
output_dir = os.path.abspath(output_dir)

set_environment(args.seed, args.cuda)
log_path = args.log_file
logger =  create_logger(__name__, to_disk=True, log_file=log_path)
logger.info(args.answer_opt)

tasks_config = {}
if os.path.exists(args.task_config_path):
    with open(args.task_config_path, 'r') as reader:
        tasks_config = json.loads(reader.read())

def dump(path, data):
    with open(path ,'w') as f:
        json.dump(data, f)

def main():
    logger.info('Launching the MT-DNN training')
    opt = vars(args)
    # update data dir
    opt['data_dir'] = data_dir
    batch_size = args.batch_size
    train_data_list = []
    tasks = {}
    tasks_class = {}
    nclass_list = []
    decoder_opts = []
    dropout_list = []

    for dataset in args.train_datasets:
        prefix = dataset.split('_')[0]
        if prefix in tasks: continue
        assert prefix in DATA_META
        assert prefix in DATA_TYPE
        data_type = DATA_TYPE[prefix]
        nclass = DATA_META[prefix]
        task_id = len(tasks)
        if args.mtl_opt > 0:
            task_id = tasks_class[nclass] if nclass in tasks_class else len(tasks_class)

        task_type = TASK_TYPE[prefix]
        pw_task = False
        if prefix in opt['pw_tasks']:
            pw_task = True

        dopt = generate_decoder_opt(prefix, opt['answer_opt'])
        if task_id < len(decoder_opts):
            decoder_opts[task_id] = min(decoder_opts[task_id], dopt)
        else:
            decoder_opts.append(dopt)

        if prefix not in tasks:
            tasks[prefix] = len(tasks)
            if args.mtl_opt < 1: nclass_list.append(nclass)

        if (nclass not in tasks_class):
            tasks_class[nclass] = len(tasks_class)
            if args.mtl_opt > 0: nclass_list.append(nclass)

        dropout_p = args.dropout_p
        if tasks_config and prefix in tasks_config:
            dropout_p = tasks_config[prefix]
        dropout_list.append(dropout_p)

        train_path = os.path.join(data_dir, '{}_train.json'.format(dataset))
        logger.info('Loading {} as task {}'.format(train_path, task_id))
        train_data = BatchGen(BatchGen.load(train_path, True, pairwise=pw_task, maxlen=args.max_seq_len),
                                batch_size=batch_size,
                                dropout_w=args.dropout_w,
                                gpu=args.cuda,
                                task_id=task_id,
                                maxlen=args.max_seq_len,
                                pairwise=pw_task,
                                data_type=data_type,
                                task_type=task_type)
        train_data_list.append(train_data)

    opt['answer_opt'] = decoder_opts
    opt['tasks_dropout_p'] = dropout_list

    args.label_size = ','.join([str(l) for l in nclass_list])
    logger.info(args.label_size)
    dev_data_list = []
    test_data_list = []
    for dataset in args.test_datasets:
        prefix = dataset.split('_')[0]
        task_id = tasks_class[DATA_META[prefix]] if args.mtl_opt > 0 else tasks[prefix]
        task_type = TASK_TYPE[prefix]

        pw_task = False
        if prefix in opt['pw_tasks']:
            pw_task = True

        assert prefix in DATA_TYPE
        data_type = DATA_TYPE[prefix]

        dev_path = os.path.join(data_dir, '{}_dev.json'.format(dataset))
        dev_data = None
        if os.path.exists(dev_path):
            dev_data = BatchGen(BatchGen.load(dev_path, False, pairwise=pw_task, maxlen=args.max_seq_len),
                                  batch_size=args.batch_size_eval,
                                  gpu=args.cuda, is_train=False,
                                  task_id=task_id,
                                  maxlen=args.max_seq_len,
                                  pairwise=pw_task,
                                  data_type=data_type,
                                  task_type=task_type)
        dev_data_list.append(dev_data)

        test_path = os.path.join(data_dir, '{}_test.json'.format(dataset))
        test_data = None
        if os.path.exists(test_path):
            test_data = BatchGen(BatchGen.load(test_path, False, pairwise=pw_task, maxlen=args.max_seq_len),
                                  batch_size=args.batch_size_eval,
                                  gpu=args.cuda, is_train=False,
                                  task_id=task_id,
                                  maxlen=args.max_seq_len,
                                  pairwise=pw_task,
                                  data_type=data_type,
                                  task_type=task_type)
        test_data_list.append(test_data)

    logger.info('#' * 20)
    logger.info(opt)
    logger.info('#' * 20)

    all_iters =[iter(item) for item in train_data_list]
    all_lens = [len(bg) for bg in train_data_list]
    num_all_batches = args.epochs * sum(all_lens)

    if len(train_data_list)> 1 and args.ratio > 0:
        num_all_batches = int(args.epochs * (len(train_data_list[0]) * (1 + args.ratio)))

    model_path = args.init_checkpoint
    state_dict = None

    if os.path.exists(model_path):
        state_dict = torch.load(model_path)
        config = state_dict['config']
        config['attention_probs_dropout_prob'] = args.bert_dropout_p
        config['hidden_dropout_prob'] = args.bert_dropout_p
        opt.update(config)
    else:
        logger.error('#' * 20)
        logger.error('Could not find the init model!\n The parameters will be initialized randomly!')
        logger.error('#' * 20)
        config = BertConfig(vocab_size_or_config_json_file=30522).to_dict()
        opt.update(config)

    model = MTDNNModel(opt, state_dict=state_dict, num_train_step=num_all_batches)
    ####model meta str
    headline = '############# Model Arch of MT-DNN #############'
    ###print network
    logger.info('\n{}\n{}\n'.format(headline, model.network))

    # dump config
    config_file = os.path.join(output_dir, 'config.json')
    with open(config_file, 'w', encoding='utf-8') as writer:
        writer.write('{}\n'.format(json.dumps(opt)))
        writer.write('\n{}\n{}\n'.format(headline, model.network))

    logger.info("Total number of params: {}".format(model.total_param))

    if args.freeze_layers > 0:
        model.network.freeze_layers(args.freeze_layers)

    if args.cuda:
        model.cuda()
    for epoch in range(0, args.epochs):
        logger.warning('At epoch {}'.format(epoch))
        for train_data in train_data_list:
            train_data.reset()
        start = datetime.now()
        all_indices=[]
        if len(train_data_list)> 1 and args.ratio > 0:
            main_indices =[0] * len(train_data_list[0])
            extra_indices=[]
            for i in range(1, len(train_data_list)):
                extra_indices += [i] * len(train_data_list[i])
            random_picks=int(min(len(train_data_list[0]) * args.ratio, len(extra_indices)))
            extra_indices = np.random.choice(extra_indices, random_picks, replace=False)
            if args.mix_opt > 0:
                extra_indices = extra_indices.tolist()
                random.shuffle(extra_indices)
                all_indices = extra_indices + main_indices
            else:
                all_indices = main_indices + extra_indices.tolist()

        else:
            for i in range(1, len(train_data_list)):
                all_indices += [i] * len(train_data_list[i])
            if args.mix_opt > 0:
                random.shuffle(all_indices)
            all_indices += [0] * len(train_data_list[0])
        if args.mix_opt < 1:
            random.shuffle(all_indices)

        for i in range(len(all_indices)):
            task_id = all_indices[i]
            batch_meta, batch_data= next(all_iters[task_id])
            model.update(batch_meta, batch_data)
            if (model.updates) % args.log_per_updates == 0 or model.updates == 1:
                logger.info('Task [{0:2}] updates[{1:6}] train loss[{2:.5f}] remaining[{3}]'.format(task_id,
                    model.updates, model.train_loss.avg,
                    str((datetime.now() - start) / (i + 1) * (len(all_indices) - i - 1)).split('.')[0]))

        for idx, dataset in enumerate(args.test_datasets):
            prefix = dataset.split('_')[0]
            label_dict = GLOBAL_MAP.get(prefix, None)
            dev_data = dev_data_list[idx]
            if dev_data is not None:
                dev_metrics, dev_predictions, scores, golds, dev_ids= eval_model(model, dev_data, dataset=prefix,
                                                                                 use_cuda=args.cuda)
                for key, val in dev_metrics.items():
                    logger.warning("Task {0} -- epoch {1} -- Dev {2}: {3:.3f}".format(dataset, epoch, key, val))
                score_file = os.path.join(output_dir, '{}_dev_scores_{}.json'.format(dataset, epoch))
                results = {'metrics': dev_metrics, 'predictions': dev_predictions, 'uids': dev_ids, 'scores': scores}
                dump(score_file, results)
                official_score_file = os.path.join(output_dir, '{}_dev_scores_{}.tsv'.format(dataset, epoch))
                submit(official_score_file, results, label_dict)

            # test eval
            test_data = test_data_list[idx]
            if test_data is not None:
                test_metrics, test_predictions, scores, golds, test_ids= eval_model(model, test_data, dataset=prefix,
                                                                                 use_cuda=args.cuda, with_label=False)
                score_file = os.path.join(output_dir, '{}_test_scores_{}.json'.format(dataset, epoch))
                results = {'metrics': test_metrics, 'predictions': test_predictions, 'uids': test_ids, 'scores': scores}
                dump(score_file, results)
                official_score_file = os.path.join(output_dir, '{}_test_scores_{}.tsv'.format(dataset, epoch))
                submit(official_score_file, results, label_dict)
                logger.info('[new test scores saved.]')

        model_file = os.path.join(output_dir, 'model_{}.pt'.format(epoch))
        model.save(model_file)


if __name__ == '__main__':
    main()



import json
import os

from setuptools import find_packages, setup

containing_dir = os.path.split(__file__)[0]

with open(os.path.join(containing_dir, "meta.json"), "r") as f:
    meta = json.load(f)

readme_path = os.path.join(containing_dir, "README.md")

with open(readme_path, "r") as f:
    long_description = f.read()

setup(
    name=meta["name"],
    author=meta["author"],
    maintainer=meta["maintainer"],
    version=meta["version"],
    packages=find_packages(exclude=["test"]),
    license="LICENSE.txt",
    description=meta["description"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Needed to ensure Dockerfiles, etc are copied with the package
    include_package_data=True,
    url=meta["url"],
    install_requires=[
        "docker >= 3.7.0",
        "numpy >= 1.16.0",
        "pandas >= 0.24.0",
        "scikit-learn >= 0.22.0",
        "ray >= 0.8.4",
        "altair >= 3.2.0",
        "click >= 7.0",
        "humanize >= 1.1.0",
    ],
    extras_require={
        "augment": ["nltk>=3.4.4", "gensim>=3.8.2", "spacy>=2.1.4"],
        "tokenize": ["sentencepiece >= 0.1.83, < 0.1.90"],
        "interactive": [
            "streamlit >= 0.56.0",
            "gensim >= 3.8.2",
            "umap-learn >= 0.3.10",
            "eli5 >= 0.10.1",
            "hdbscan >= 0.8.24",
        ],
    },
    python_requires=">=3.7",
    entry_points={"console_scripts": ["gobbli = gobbli.cli:main"]},
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
)



# Copyright (c) Microsoft. All rights reserved.
import random
import torch
import numpy
from torch.autograd import Variable
import subprocess

class AverageMeter(object):
    """Computes and stores the average and current value."""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def set_environment(seed, set_cuda=False):
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available() and set_cuda:
        torch.cuda.manual_seed_all(seed)

def patch_var(v, cuda=True):
    if cuda:
        v = Variable(v.cuda(async=True))
    else:
        v = Variable(v)
    return v

def get_gpu_memory_map():
    result = subprocess.check_output(
        [
            'nvidia-smi', '--query-gpu=memory.used',
            '--format=csv,nounits,noheader'
        ], encoding='utf-8')
    gpu_memory = [int(x) for x in result.strip().split('\n')]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map

def get_pip_env():
    result = subprocess.call(["pip", "freeze"])
    return result



import json
import shutil
import tempfile
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd

import gobbli.io
from gobbli.docker import maybe_mount, run_container
from gobbli.model.base import BaseModel
from gobbli.model.context import ContainerTaskContext
from gobbli.model.mixin import PredictMixin, TrainMixin
from gobbli.util import (
    assert_in,
    assert_type,
    copy_file,
    download_file,
    escape_line_delimited_texts,
)

MTDNN_MODEL_FILES = {
    "mt-dnn-base": "https://mrc.blob.core.windows.net/mt-dnn-model/mt_dnn_base.pt",
    "mt-dnn-large": "https://mrc.blob.core.windows.net/mt-dnn-model/mt_dnn_large.pt",
}
"""
A mapping from model names to weight files.
"mt-dnn-base" is a safe default for most situations.
Larger models require more time and GPU memory to run.
"""


def _preprocess_text(text_series: pd.Series) -> pd.Series:
    """
    Preprocess a Series of text for the MT-DNN uncased model.
    """
    return text_series.str.lower().str.replace(r"\s", " ", regex=True)


def _df_to_tsv(df: pd.DataFrame, output_file: Path):
    """
    Write a dataframe with "X" and "y" (optional) columns to the given
    output file in the format expected by MT-DNN for a TSV file.
    """
    df.loc[:, "X"] = _preprocess_text(df["X"])
    df.to_csv(output_file, index=False, header=True)


def _write_labels(labels: List[Any], output_file: Path):
    """
    Write the given set of labels to the given file.
    """
    output_file.write_text(escape_line_delimited_texts(labels))


class MTDNN(BaseModel, TrainMixin, PredictMixin):
    """
    Classifier wrapper for Microsoft's MT-DNN:
    https://github.com/namisan/mt-dnn
    """

    _BUILD_PATH = Path(__file__).parent

    _TRAIN_INPUT_FILE = "train.csv"
    _VALID_INPUT_FILE = "valid.csv"
    _TEST_INPUT_FILE = "test.csv"
    _LABELS_INPUT_FILE = "labels.csv"

    _PREDICT_OUTPUT_FILE = "predict.csv"

    _LOG_FILE = "log.log"

    _WEIGHTS_FILE_NAME = "weights.pt"

    def init(self, params: Dict[str, Any]):
        """
        See :meth:`gobbli.model.base.BaseModel.init`.

        MT-DNN parameters:

        - ``max_seq_length`` (:obj:`int`): The maximum total input sequence length after
          WordPiece tokenization.  Sequences longer than this will be truncated,
          and sequences shorter than this will be padded.  Default: 128
        - ``mtdnn_model`` (:obj:`str`): Name of a pretrained MT-DNN model to use.
          See :obj:`MTDNN_MODEL_FILES` for a listing of available MT-DNN models.
        """
        self.max_seq_length = 128
        self.mtdnn_model = "mt-dnn-base"

        for name, value in params.items():
            if name == "max_seq_length":
                assert_type(name, value, int)
                self.max_seq_length = value
            elif name == "mtdnn_model":
                assert_in(name, value, set(MTDNN_MODEL_FILES.keys()))
                self.mtdnn_model = value
            else:
                raise ValueError(f"Unknown param '{name}'")

    @property
    def weights_dir(self) -> Path:
        """
        Returns:
          The directory containing pretrained weights for this instance.
        """
        return self.class_weights_dir / self.mtdnn_model

    @property
    def image_tag(self) -> str:
        """
        Returns:
          The Docker image tag to be used for the MT-DNN container.
        """
        return "gobbli-mt-dnn-classifier"

    def _build(self):
        # Download data if we don't already have it
        if not self.weights_dir.exists():
            with tempfile.TemporaryDirectory() as tmpdir:
                tmp_weights_dir = Path(tmpdir) / self.weights_dir.name
                tmp_weights_dir.mkdir()
                self.logger.info("Downloading pre-trained weights.")
                downloaded_file = download_file(MTDNN_MODEL_FILES[self.mtdnn_model])
                copy_file(downloaded_file, tmp_weights_dir / MTDNN._WEIGHTS_FILE_NAME)
                shutil.move(tmp_weights_dir, self.weights_dir)
                self.logger.info("Weights downloaded.")

        # Build the docker image
        self.docker_client.images.build(
            path=str(MTDNN._BUILD_PATH),
            tag=self.image_tag,
            **self._base_docker_build_kwargs,
        )

    @staticmethod
    def _get_checkpoint(
        user_checkpoint: Optional[Path], context: ContainerTaskContext
    ) -> Tuple[Optional[Path], Path]:
        """
        Determines the host checkpoint file and container checkpoint file
        using the user-requested checkpoint (if any) and the container context.

        Args:
          user_checkpoint: An optional checkpoint passed in by the user.  If the user doesn't
            pass one, use the default pretrained checkpoint.
          context: The container context to create the checkpoint in.

        Returns:
          A 2-tuple: the host checkpoint file (if any) and
            the container checkpoint file
        """
        if user_checkpoint is None:
            # Default weights
            host_checkpoint_file = None
            container_checkpoint_file = (
                BaseModel._CONTAINER_WEIGHTS_PATH / MTDNN._WEIGHTS_FILE_NAME
            )
        else:
            # Trained weights, which will be mounted in the container
            host_checkpoint_file = user_checkpoint
            container_checkpoint_file = context.container_root_dir / "checkpoint.pt"

        return host_checkpoint_file, container_checkpoint_file

    def _write_input(self, X: List[str], y: Optional[List[str]], input_file: Path):
        """
        Write the given gobbli input into the format expected by MT-DNN.
        Make sure the given directory exists first.
        """
        df_data = {"X": X}
        if y is not None:
            df_data["y"] = y

        df = pd.DataFrame(df_data)

        _df_to_tsv(df, input_file)

    def _train(
        self, train_input: gobbli.io.TrainInput, context: ContainerTaskContext
    ) -> gobbli.io.TrainOutput:
        if train_input.multilabel:
            raise ValueError(
                "gobbli MT-DNN model doesn't support multilabel classification."
            )

        self._write_input(
            train_input.X_train,
            train_input.y_train_multiclass,
            context.host_input_dir / MTDNN._TRAIN_INPUT_FILE,
        )
        self._write_input(
            train_input.X_valid,
            train_input.y_valid_multiclass,
            context.host_input_dir / MTDNN._VALID_INPUT_FILE,
        )

        labels = train_input.labels()
        labels_path = context.host_input_dir / MTDNN._LABELS_INPUT_FILE
        _write_labels(labels, labels_path)

        if train_input.valid_batch_size != train_input.train_batch_size:
            warnings.warn(
                "MT-DNN model does not support separate validation batch size; "
                f"using train batch size '{train_input.train_batch_size}' for both "
                "training and validation."
            )

        # Determine checkpoint to use
        host_checkpoint_file, container_checkpoint_file = self._get_checkpoint(
            train_input.checkpoint, context
        )

        cmd = (
            "python gobbli_train.py"
            " --data_dir=data/mt_dnn"
            f" --init_checkpoint={container_checkpoint_file}"
            f" --batch_size={train_input.train_batch_size}"
            f" --output_dir={context.container_output_dir}"
            f" --log_file={context.container_output_dir / MTDNN._LOG_FILE}"
            " --optimizer=adamax"
            " --grad_clipping=0"
            " --global_grad_clipping=1"
            " --lr=2e-5"
            f" --train_file={context.container_input_dir / MTDNN._TRAIN_INPUT_FILE}"
            f" --valid_file={context.container_input_dir / MTDNN._VALID_INPUT_FILE}"
            f" --label_file={context.container_input_dir / MTDNN._LABELS_INPUT_FILE}"
            f" --epochs={train_input.num_train_epochs}"
            f" --max_seq_len={self.max_seq_length}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_file, container_checkpoint_file
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # MT-DNN counts epochs starting from 0
        final_epoch = train_input.num_train_epochs - 1

        # Parse the generated evaluation results files
        eval_results = {}  # type: Dict[str, Any]
        for name in ("train", "valid"):
            results_file = context.host_output_dir / f"{name}_scores_{final_epoch}.json"
            with open(results_file, "r") as f:
                results = json.load(f)
                eval_results.update(
                    {f"{name}_{key}": val for key, val in results.items()}
                )

        return gobbli.io.TrainOutput(
            valid_loss=eval_results["valid_metrics"]["loss"],
            valid_accuracy=eval_results["valid_metrics"]["accuracy"] / 100,
            train_loss=eval_results["train_metrics"]["loss"],
            labels=labels,
            multilabel=False,
            checkpoint=context.host_output_dir / f"model_{final_epoch}.pt",
            _console_output=container_logs,
        )

    def _predict(
        self, predict_input: gobbli.io.PredictInput, context: ContainerTaskContext
    ) -> gobbli.io.PredictOutput:
        self._write_input(
            predict_input.X, None, context.host_input_dir / MTDNN._TEST_INPUT_FILE
        )

        labels_path = context.host_input_dir / MTDNN._LABELS_INPUT_FILE
        _write_labels(predict_input.labels, labels_path)

        # Determine checkpoint to use
        host_checkpoint_file, container_checkpoint_file = self._get_checkpoint(
            predict_input.checkpoint, context
        )

        cmd = (
            "python gobbli_train.py"
            " --data_dir=data/mt_dnn"
            f" --init_checkpoint={container_checkpoint_file}"
            f" --batch_size={predict_input.predict_batch_size}"
            f" --output_dir={context.container_output_dir}"
            f" --log_file={context.container_output_dir / MTDNN._LOG_FILE}"
            " --optimizer=adamax"
            " --grad_clipping=0"
            " --global_grad_clipping=1"
            " --lr=2e-5"
            f" --test_file={context.container_input_dir / MTDNN._TEST_INPUT_FILE}"
            f" --label_file={context.container_input_dir / MTDNN._LABELS_INPUT_FILE}"
            f" --max_seq_len={self.max_seq_length}"
        )

        run_kwargs = self._base_docker_run_kwargs(context)

        # Mount the checkpoint in the container if needed
        maybe_mount(
            run_kwargs["volumes"], host_checkpoint_file, container_checkpoint_file
        )

        container_logs = run_container(
            self.docker_client, self.image_tag, cmd, self.logger, **run_kwargs
        )

        # Retrieve the generated predictions
        return gobbli.io.PredictOutput(
            y_pred_proba=pd.read_csv(
                context.host_output_dir / MTDNN._PREDICT_OUTPUT_FILE
            ),
            _console_output=container_logs,
        )



import functools
import random
from typing import Any, List

from gobbli.augment.base import BaseAugment


def _detokenize_doc(doc: Any) -> str:
    """
    Detokenize a spaCy Doc object back into a string, applying our custom replacements
    as needed. This requires the associated extension to have been registered appropriately.
    The :class:`WordNet` constructor should handle registering the extension.
    """
    return "".join([f"{tok._.replacement}{tok.whitespace_}" for tok in doc])


def _get_lemmas(synsets: List[Any]) -> List[str]:
    """
    Return all the lemma names associated with a list of synsets.
    """
    return [lemma_name for synset in synsets for lemma_name in synset.lemma_names()]


@functools.lru_cache(maxsize=256)
def _get_wordnet_lemmas(word: str, pos: str) -> List[str]:
    """
    Determine all the lemmas for a given word to be considered candidates for
    replacement.  Wrap this function in an LRU cache to keep from recalculating common words
    or terms reused frequently in the same document.
    """
    # We should have properly guarded this import in the WordNet constructor
    from nltk.corpus import wordnet

    synsets = wordnet.synsets(word, pos)
    hypernyms = [hypernym for synset in synsets for hypernym in synset.hypernyms()]
    hyponyms = [hyponym for synset in synsets for hyponym in synset.hyponyms()]
    return list(
        frozenset(_get_lemmas(synsets) + _get_lemmas(hypernyms) + _get_lemmas(hyponyms))
    )


class WordNet(BaseAugment):
    """
    Data augmentation method based on WordNet.  Replaces words with similar
    words according to the WordNet ontology.  Texts will be Part of Speech-tagged
    using spaCy to help ensure only sensible replacements (i.e., within the same part of speech)
    are considered.

    Args:
      skip_download_check: If True, don't try to download the WordNet corpus;
        assume it's already been downloaded.
      spacy_model: The language model to be used for Part of Speech tagging by spaCy.
        The model must already have been installed.
    """

    def __init__(self, skip_download_check: bool = False, spacy_model="en_core_web_sm"):
        try:
            from nltk.corpus import wordnet
            import nltk
        except ImportError:
            raise ImportError(
                "WordNet-based data augmentation requires nltk to be installed."
            )

        self.wn = wordnet

        try:
            import spacy
            from spacy.tokens import Token
        except ImportError:
            raise ImportError(
                "WordNet-based data augmentation requires spaCy and a language "
                "model to be installed (for part of speech tagging)."
            )

        if not skip_download_check:
            nltk.download("wordnet")

        self.nlp = spacy.load(spacy_model, parser=False, tagger=True, entity=False)
        Token.set_extension("replacement", default=None, force=True)

    def _maybe_replace_token(self, token: Any) -> str:
        if token.pos_ == "ADJ":
            wordnet_pos = self.wn.ADJ
        elif token.pos_ == "NOUN":
            wordnet_pos = self.wn.NOUN
        elif token.pos_ == "VERB":
            wordnet_pos = self.wn.VERB
        elif token.pos_ == "ADV":
            wordnet_pos = self.wn.ADV
        else:
            # If the token's part of speech isn't recognized by WordNet,
            # return it without replacing.
            return token.text

        all_lemma_names = _get_wordnet_lemmas(token.text, wordnet_pos)
        if len(all_lemma_names) > 0:
            # WordNet lemmas have underscores where spaces should be, so apply spaces
            # appropriately
            return random.choice(all_lemma_names).replace("_", " ")
        else:
            return token.text

    def augment(self, X: List[str], times: int = 5, p: float = 0.1) -> List[str]:
        new_texts = []
        tagged_docs = [doc for doc in self.nlp.pipe(X)]

        for _ in range(times):
            for doc in tagged_docs:
                for tok in doc:
                    if random.random() < p:
                        tok._.replacement = self._maybe_replace_token(tok)
                    else:
                        tok._.replacement = tok.text

                new_texts.append(_detokenize_doc(doc))
        return new_texts



# Copyright (c) Microsoft. All rights reserved.
import numpy as np
from random import shuffle
from .label_map import METRIC_FUNC, METRIC_META, METRIC_NAME

def load_scitail(file, label_dict):
    """Loading data of scitail
    """
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            blocks = line.strip().split('\t')
            assert len(blocks) > 2
            if blocks[0] == '-': continue
            sample = {'uid': str(cnt), 'premise': blocks[0], 'hypothesis': blocks[1], 'label': label_dict[blocks[2]]}
            rows.append(sample)
            cnt += 1
    return rows

def load_snli(file, label_dict, header=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            assert len(blocks) > 10
            if blocks[-1] == '-': continue
            lab = label_dict[blocks[-1]]
            if lab is None:
                import pdb; pdb.set_trace()
            lab = 0 if lab is None else lab
            sample = {'uid': blocks[0], 'premise': blocks[7], 'hypothesis': blocks[8], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_mnli(file, label_dict, header=True, multi_snli=False, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            assert len(blocks) > 9
            if blocks[-1] == '-': continue
            lab = 0
            if is_train:
                lab = label_dict[blocks[-1]]
            if lab is None:
                import pdb; pdb.set_trace()
            lab = 0 if lab is None else lab
            sample = {'uid': blocks[0], 'premise': blocks[8], 'hypothesis': blocks[9], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_mrpc(file, header=True, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            assert len(blocks) > 4
            lab = 0
            if is_train:
                lab = int(blocks[0])
            sample = {'uid': cnt, 'premise': blocks[-2], 'hypothesis': blocks[-1], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_qnli(file, label_dict, header=True, is_train=True):
    """QNLI for classification"""
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            assert len(blocks) > 2
            lab = 0
            if is_train:
                lab = label_dict[blocks[-1]]
            if lab is None:
                import pdb; pdb.set_trace()
            lab = 0 if lab is None else lab
            sample = {'uid': blocks[0], 'premise': blocks[1], 'hypothesis': blocks[2], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_qnnli(file, label_dict, header=True, is_train=True):
    """QNLI for ranking"""
    rows = []
    mis_matched_cnt = 0
    cnt = 0
    with open(file, encoding="utf8") as f:
        lines = f.readlines()
        if header: lines = lines[1:]

        assert len(lines) % 2 == 0
        for idx in range(0, len(lines), 2):
            block1 = lines[idx].strip().split('\t')
            block2 = lines[idx + 1].strip().split('\t')
            # train shuffle
            assert len(block1) > 2 and len(block2) > 2
            if is_train and block1[1] != block2[1]:
                mis_matched_cnt += 1
                continue
            assert block1[1] == block2[1]
            lab1, lab2 = 0, 0
            if is_train:
                blocks = [block1, block2]
                shuffle(blocks)
                block1 = blocks[0]
                block2 = blocks[1]
                lab1 = label_dict[block1[-1]]
                lab2 = label_dict[block2[-1]]
                if lab1 == lab2:
                    mis_matched_cnt += 1
                    continue
            lab = int(np.argmax([lab1, lab2]))
            sample = {'uid': cnt, 'premise': block1[1], 'hypothesis': [block1[2], block2[2]], 'label': lab, 'ruid':[block1[0], block2[0]], 'olabel':[lab1, lab2]}
            cnt += 1
            rows.append(sample)
    return rows

def load_qqp(file, header=True, is_train=True):
    rows = []
    cnt = 0
    skipped = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            if is_train and len(blocks) < 6:
                skipped += 1
                continue
            if not is_train: assert len(blocks) == 3
            lab = 0
            if is_train:
                lab = int(blocks[-1])
                sample = {'uid': cnt, 'premise': blocks[-3], 'hypothesis': blocks[-2], 'label': lab}
            else:
                sample = {'uid': int(blocks[0]), 'premise': blocks[-2], 'hypothesis': blocks[-1], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_rte(file, label_dict, header=True, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header =False
                continue
            blocks = line.strip().split('\t')
            if is_train and len(blocks) < 4: continue
            if not is_train: assert len(blocks) == 3
            lab = 0
            if is_train:
                lab = label_dict[blocks[-1]]
                sample = {'uid': int(blocks[0]), 'premise': blocks[-3], 'hypothesis': blocks[-2], 'label': lab}
            else:
                sample = {'uid': int(blocks[0]), 'premise': blocks[-2], 'hypothesis': blocks[-1], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_wnli(file, header=True, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header =False
                continue
            blocks = line.strip().split('\t')
            if is_train and len(blocks) < 4: continue
            if not is_train: assert len(blocks) == 3
            lab = 0
            if is_train:
                lab = int(blocks[-1])
            sample = {'uid': cnt, 'premise': blocks[-2], 'hypothesis': blocks[-1], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_diag(file, label_dict, header=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            assert len(blocks) > 3
            sample = {'uid': cnt, 'premise': blocks[-3], 'hypothesis': blocks[-2], 'label': label_dict[blocks[-1]]}
            rows.append(sample)
            cnt += 1
    return rows

def load_sst(file, header=True, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            if is_train and len(blocks) < 2: continue
            lab = 0
            if is_train:
                lab = int(blocks[-1])
                sample = {'uid': cnt, 'premise': blocks[0], 'label': lab}
            else:
                sample = {'uid': int(blocks[0]), 'premise': blocks[1], 'label': lab}

            cnt += 1
            rows.append(sample)
    return rows

def load_cola(file, header=True, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            if is_train and len(blocks) < 2: continue
            lab = 0
            if is_train:
                lab = int(blocks[1])
                sample = {'uid': cnt, 'premise': blocks[-1], 'label': lab}
            else:
                sample = {'uid': cnt, 'premise': blocks[-1], 'label': lab}
            rows.append(sample)
            cnt += 1
    return rows

def load_sts(file, header=True, is_train=True):
    rows = []
    cnt = 0
    with open(file, encoding="utf8") as f:
        for line in f:
            if header:
                header = False
                continue
            blocks = line.strip().split('\t')
            assert len(blocks) > 8
            score = 0.0
            if is_train:
                score = float(blocks[-1])
                sample = {'uid': cnt, 'premise': blocks[-3],'hypothesis': blocks[-2], 'label': score}
            else:
                sample = {'uid': cnt, 'premise': blocks[-2],'hypothesis': blocks[-1], 'label': score}
            rows.append(sample)
            cnt += 1
    return rows

def submit(path, data, label_dict=None):
    header = 'index\tprediction'
    with open(path ,'w') as writer:
        predictions, uids = data['predictions'], data['uids']
        writer.write('{}\n'.format(header))
        assert len(predictions) == len(uids)
        # sort label
        paired = [(int(uid), predictions[idx]) for idx, uid in enumerate(uids)]
        paired = sorted(paired, key=lambda item: item[0])
        for uid, pred in paired:
            if label_dict is None:
                writer.write('{}\t{}\n'.format(uid, pred))
            else:
                assert type(pred) is int
                writer.write('{}\t{}\n'.format(uid, label_dict[pred]))

def _truncate_seq_pair(tokens_a, tokens_b, max_length):
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()

def eval_model(model, data, dataset, use_cuda=True, with_label=True):
    data.reset()
    if use_cuda:
        model.cuda()
    predictions = []
    golds = []
    scores = []
    ids = []
    metrics = {}
    for batch_meta, batch_data in data:
        score, pred, gold = model.predict(batch_meta, batch_data)
        predictions.extend(pred)
        golds.extend(gold)
        scores.extend(score)
        ids.extend(batch_meta['uids'])
    mmeta = METRIC_META[dataset]
    if with_label:
        for mm in mmeta:
            metric_name = METRIC_NAME[mm]
            metric_func = METRIC_FUNC[mm]
            if mm < 3:
                metric = metric_func(predictions, golds)
            else:
                metric = metric_func(scores, golds)
            metrics[metric_name] = metric
    return metrics, predictions, scores, golds, ids



from pathlib import Path
from typing import Set, Tuple

from gobbli.dataset.nested_file import NestedFileDataset
from gobbli.util import download_archive


class NewsgroupsDataset(NestedFileDataset):
    """
    gobbli Dataset for the 20 Newsgroups problem.

    http://qwone.com/~jason/20Newsgroups/
    """

    def labels(self) -> Set[str]:
        return {
            "alt.atheism",
            "comp.graphics",
            "comp.os.ms-windows.misc",
            "comp.sys.ibm.pc.hardware",
            "comp.sys.mac.hardware",
            "comp.windows.x",
            "misc.forsale",
            "rec.autos",
            "rec.motorcycles",
            "rec.sport.baseball",
            "rec.sport.hockey",
            "sci.crypt",
            "sci.electronics",
            "sci.med",
            "sci.space",
            "soc.religion.christian",
            "talk.politics.guns",
            "talk.politics.mideast",
            "talk.politics.misc",
            "talk.religion.misc",
        }

    def download(self, data_dir: Path) -> Path:
        return download_archive(
            "https://ndownloader.figshare.com/files/5975967",
            data_dir,
            filename="20news-bydate.tar.gz",
        )

    def folders(self) -> Tuple[Path, Path]:
        return Path("20news-bydate-train"), Path("20news-bydate-test")

    def read_source_file(self, file_path: Path) -> str:
        return file_path.read_text(encoding="latin-1")



import argparse

import torch
import torch.nn.functional as F
from transformers import BertConfig, BertForMaskedLM, BertTokenizer


def batch_list(l, batch_size):
    for i in range(0, len(l), batch_size):
        yield l[i : i + batch_size]


def encode_batch(batch, tokenizer, config):
    # Truncate input based on the sequence length
    # before stacking as a batch
    # Also return a boolean array indicating which tokens are masked
    encoded_texts = []
    for text in batch:
        encoded_text = torch.tensor(tokenizer.encode(text))[
            : config.max_position_embeddings
        ]
        encoded_texts.append(encoded_text)

    return torch.nn.utils.rnn.pad_sequence(encoded_texts, batch_first=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "input_file", help="Input file containing line-delimited texts."
    )
    parser.add_argument(
        "output_file",
        help="Output file, where line-delimited generated texts will be written.",
    )
    parser.add_argument(
        "--bert-model",
        help="BERT model to use. "
        "Can be one of the pretrained names supported by transformers, in which case "
        "the pretrained weights will be downloaded. "
        "Anything else supported by transformers should work as well. ",
        default="bert-base-uncased",
    )
    parser.add_argument(
        "--n-probable",
        help="Number of probable tokens to consider for replacement.",
        type=int,
        default=5,
    )
    parser.add_argument(
        "--diversity",
        help="Inverse dependence of selection likelihood on predicted probability.",
        type=float,
        default=0.8,
    )
    parser.add_argument(
        "--probability",
        help="Probability of masking each token.",
        type=float,
        default=0.1,
    )
    parser.add_argument(
        "--times",
        help="Number of new documents to generate for each existing document.",
        type=int,
        default=5,
    )
    parser.add_argument(
        "--batch-size",
        help="Number of documents to run through the BERT model at once.",
        type=int,
        default=32,
    )
    parser.add_argument(
        "--cache-dir",
        help="Directory to use as the cache for transformers downloads.",
        default=None,
    )
    parser.add_argument(
        "--device",
        default="cpu",
        help="PyTorch name for the device to use when running the model. Default: %(default)s",
    )

    args = parser.parse_args()

    if args.batch_size < 1:
        raise ValueError("batch_size must be >= 1")

    if args.times < 0:
        raise ValueError("times must be >= 0")

    if not 0 <= args.probability <= 1:
        raise ValueError("probability must be >= 0 and <= 1")

    if not 0 < args.diversity <= 1:
        raise ValueError("diversity must be > 0 and <= 1")

    tokenizer = BertTokenizer.from_pretrained(args.bert_model, cache_dir=args.cache_dir)
    if tokenizer is None:
        raise ValueError("Failed to acquire tokenizer")
    config = BertConfig.from_pretrained(args.bert_model, cache_dir=args.cache_dir)
    if config is None:
        raise ValueError("Failed to acquire config")
    model = BertForMaskedLM.from_pretrained(
        args.bert_model, config=config, cache_dir=args.cache_dir
    )
    if model is None:
        raise ValueError("Failed to acquire model")

    model = model.to(args.device)
    model.eval()

    with open(args.input_file, "r", encoding="utf-8") as f_in:
        with open(args.output_file, "w", encoding="utf-8") as f_out:
            batches = batch_list(f_in.readlines(), args.batch_size)
            input_id_batches = [
                encode_batch(batch, tokenizer, config) for batch in batches
            ]

            for time in range(args.times):
                for batch_id, input_ids in enumerate(input_id_batches):
                    # Generate a replacement mask for the batch.  Do this each time we
                    # run the batch to get different results
                    # Don't mask any padding tokens, so we don't get predictions for them
                    # and can easily remove them at the end
                    should_replace = (
                        torch.rand_like(input_ids, dtype=torch.float) < args.probability
                    ) & (input_ids != tokenizer.vocab.get(tokenizer.pad_token))

                    masked_ids = input_ids.clone().detach()
                    masked_ids[should_replace] = tokenizer.vocab.get(
                        tokenizer.mask_token
                    )
                    masked_ids = masked_ids.to(args.device)

                    # These are created implicitly, but we need to explicitly
                    # initialize them to their defaults because they won't
                    # be on the chosen device otherwise
                    attention_mask = torch.ones_like(masked_ids).to(args.device)
                    token_type_ids = torch.zeros_like(masked_ids).to(args.device)

                    with torch.no_grad():
                        (output,) = model(
                            masked_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids,
                        )

                    # Bring model results back to the CPU for processing
                    masked_ids = masked_ids.cpu()
                    output = output.cpu()

                    # Exponentiate according to inverse diversity to set level of
                    # dependence on predicted probability
                    prediction_scores = torch.pow(
                        F.softmax(output, dim=2), 1.0 / args.diversity
                    )

                    max_seq_len = prediction_scores.size(1)

                    output_ids = []
                    for i, row in enumerate(prediction_scores):
                        # size: (seq_len, vocab_len)
                        # Generate probabilities for each token.  Ultimately, we only
                        # care about the ones that were masked
                        # First generate the indices of all candidates, i.e.
                        # tokens in the top n_probable when sorted by descending predicted
                        # probability
                        # size: (seq_len, n_probable)
                        candidate_ndxs = torch.argsort(row, dim=1, descending=True)[
                            :, : args.n_probable
                        ]

                        candidates = torch.stack(
                            [tok[candidate_ndxs[i]] for i, tok in enumerate(row)]
                        )

                        # Determine a replacement among the candidates for each token
                        # in the row
                        # size: (seq_len)
                        replacement_sorted_ndxs = torch.multinomial(
                            candidates, 1
                        ).squeeze()

                        # Map the candidate indices back to original row indices
                        # size: (seq_len)
                        replacements = torch.tensor(
                            [
                                candidate_ndxs[i][ndx]
                                for i, ndx in enumerate(replacement_sorted_ndxs)
                            ]
                        )

                        # Perform the replacement
                        output_ids.append(
                            torch.where(
                                should_replace[i], replacements, masked_ids[i]
                            ).tolist()
                        )

                    # Convert the token IDs back to text; filter out padding tokens
                    output_texts = [
                        tokenizer.decode(
                            [
                                tok
                                for tok in row
                                if not tok == tokenizer.vocab.get(tokenizer.pad_token)
                            ]
                        )
                        for row in output_ids
                    ]

                    # Write text to the output file
                    # Escape any embedded newlines
                    # Make sure not to write an extra newline after the last row
                    for i, row in enumerate(output_texts):
                        f_out.write(row.replace("\n", " "))

                        # Write a newline except for the last row on the last
                        # time through.  We'd get an empty line at the end of the
                        # file otherwise
                        if not (
                            i == len(output_texts) - 1
                            and batch_id == len(input_id_batches) - 1
                            and time == args.times - 1
                        ):
                            f_out.write("\n")
